@@ -28,6 +28,7 @@
 #include <errno.h>
 #include <string.h>
 #include <syslog.h>
+#include <time.h>
 #include <sys/types.h>
 #include <sys/socket.h>
 #include <sys/utsname.h>
@@ -66,18 +67,24 @@
 #define AUTH_CRAM_MD5   1 << 2
 #define AUTH_DIGEST_MD5 1 << 3
 
+#define DIGEST_MD5_REALM_LEN   256
+#define DIGEST_MD5_NONCE_LEN   64
+#define DIGEST_MD5_CNONCE_LEN  33
+#define DIGEST_MD5_QOP_LEN     64
+#define DIGEST_MD5_URI_LEN     261
+
 #define B64(c)  (isascii(c) ? base64val[(int)(c)] : -1)
 
 extern int retry_writev(socket_t *sock, struct iovec *iov, int iovcnt);
 extern int socket_read(socket_t *sock, char *buf, size_t len);
 extern int socket_close(socket_t *sock);
 extern void socket_perror(const char *func, socket_t *sock, int ret);
+extern void set_timeout(int timeout);
 
 void md5_hex_hmac(char *hexdigest, unsigned char *text, unsigned int text_len, unsigned char *key, unsigned int key_len);
 void hmac_md5(unsigned char *text, unsigned int text_len, unsigned char *key, unsigned int key_len, unsigned char *digest);
 int start_tls(smtp_t *smtp, config_t *cfg);
 
-
 int smtp_quit(socket_t *sock, config_t *cfg);
 int auth_plain(socket_t *sock, config_t *cfg);
 int auth_login(socket_t *sock, config_t *cfg);
@@ -97,6 +104,22 @@ static const char base64val[128] = {
     41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 51, -1, -1, -1, -1, -1
 };
 
+static void
+bin2hex(char *out, const unsigned char *in, int in_len) {
+    static const char hex[17] = "0123456789abcdef";
+    int cnt;
+
+    for(cnt=0; cnt<in_len; cnt++) {
+        out[cnt * 2]       = hex[in[cnt] >> 4];
+        out[(cnt * 2) + 1] = hex[in[cnt]&0x0F];
+    }
+}
+
+void
+make_digest(char *md5str, unsigned char *digest) {
+    bin2hex(md5str, digest, 16);
+    md5str[32] = '\0';
+}
 
 void
 md5_hex_hmac(char *hexdigest, unsigned char *text, unsigned int text_len, unsigned char *key, unsigned int key_len) {
@@ -238,6 +261,181 @@ base64_decode(char *out, const char *in, int inlen) {
 }
 
 
+int
+extract_token(const char *str, const char *token, char *value, int len) {
+
+    char *p = NULL, *q = NULL;
+
+    memset(value,0x00,sizeof(char)*len);
+    if((p = strstr(str, token)) != NULL) {
+        p += strlen(token);
+        if(*p == '\"') {
+            if((q = strchr(p + 1, '\"')) == NULL)
+                return -1;
+            strncpy(value, p + 1, q - p - 1 >= len ? len - 1 : q - p - 1);
+
+	} else {
+            if((q = strchr(p, ',')) == NULL)
+                q += strlen(p);
+	    strncpy(value, p, q - p >= len ? len - 1 : q - p);
+        }
+    }
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): str=%s", str);
+#endif
+
+    return 0;
+}
+
+void
+digest_md5(char *response, unsigned char *text, unsigned int text_len, const char *login, const char *passwd) {
+
+    char realm[DIGEST_MD5_REALM_LEN];
+    char nonce[DIGEST_MD5_NONCE_LEN];
+    char qop[DIGEST_MD5_QOP_LEN];
+    char uri[DIGEST_MD5_URI_LEN];
+    char cnonce[DIGEST_MD5_CNONCE_LEN];
+
+    unsigned char random[16];
+
+    MD5_CTX ctx;
+    unsigned char digest[16];
+    char hexA1[33], hexA2[33], resp[33];
+
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): text=%s", text);
+#endif
+
+    extract_token(text, "nonce=", nonce, DIGEST_MD5_NONCE_LEN);
+    extract_token(text, "realm=", realm, DIGEST_MD5_REALM_LEN);
+    extract_token(text, "qop=", qop, DIGEST_MD5_QOP_LEN);
+
+    srand(time(NULL));
+    snprintf(random, sizeof(random), "%ld", (long int)rand());
+    bin2hex(cnonce, random, 8);
+    cnonce[16] = '\0';
+
+    sprintf(uri, "smtp/%s", realm);
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): realm=%s", realm);
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): nonce=%s", nonce);
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): qop=%s", qop);
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): cnonce=%s", cnonce);
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): uri=%s", uri);
+#endif
+
+    /* A1 */
+#ifdef USE_SSL
+    MD5_Init(&ctx);
+    MD5_Update(&ctx, login, strlen(login));
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, realm, strlen(realm));
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, passwd, strlen(passwd));
+    MD5_Final(digest, &ctx);
+#else
+    MD5Init(&ctx);
+    MD5Update(&ctx, login, strlen(login));
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, realm, strlen(realm));
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, passwd, strlen(passwd));
+    MD5Final(digest, &ctx);
+#endif
+
+#ifdef USE_SSL
+    MD5_Init(&ctx);
+    MD5_Update(&ctx, digest, 16);
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, nonce, strlen(nonce));
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, cnonce, strlen(cnonce));
+    MD5_Final(digest, &ctx);
+#else
+    MD5Init(&ctx);
+    MD5Update(&ctx, digest, 16);
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, nonce, strlen(nonce));
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, cnonce, strlen(cnonce));
+    MD5Final(digest, &ctx);
+#endif
+    make_digest(hexA1, digest);
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): A1=%s", hexA1);
+#endif
+
+    /* A2 */
+#ifdef USE_SSL
+    MD5_Init(&ctx);
+    MD5_Update(&ctx, "AUTHENTICATE:", sizeof("AUTHENTICATE:") - 1);
+    MD5_Update(&ctx, uri, strlen(uri));
+    if(!strcmp(qop, "auth-int")) {
+        MD5_Update(&ctx, ":00000000000000000000000000000000", sizeof(":00000000000000000000000000000000") - 1);
+    }
+    MD5_Final(digest, &ctx);
+#else
+    MD5Init(&ctx);
+    MD5Update(&ctx, "AUTHENTICATE:", sizeof("AUTHENTICATE:") - 1);
+    MD5Update(&ctx, uri, strlen(uri));
+    if(!strcmp(qop, "auth-int")) {
+        MD5Update(&ctx, ":00000000000000000000000000000000", sizeof(":00000000000000000000000000000000") - 1);
+    }
+    MD5Final(digest, &ctx);
+#endif
+    make_digest(hexA2, digest);
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): A2=%s", hexA2);
+#endif
+
+    /* response */
+#ifdef USE_SSL
+    MD5_Init(&ctx);
+    MD5_Update(&ctx, hexA1, 32);
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, nonce, strlen(nonce));
+    MD5_Update(&ctx, ":00000001:", sizeof(":00000001:") - 1);
+    MD5_Update(&ctx, cnonce, strlen(cnonce));
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, qop, strlen(qop));
+    MD5_Update(&ctx, ":", 1);
+    MD5_Update(&ctx, hexA2, 32);
+    MD5_Final(digest, &ctx);
+#else
+    MD5Init(&ctx);
+    MD5Update(&ctx, hexA1, 32);
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, nonce, strlen(nonce));
+    MD5Update(&ctx, ":00000001:", sizeof(":00000001:") - 1);
+    MD5Update(&ctx, cnonce, strlen(cnonce));
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, qop, strlen(qop));
+    MD5Update(&ctx, ":", 1);
+    MD5Update(&ctx, hexA2, 32);
+    MD5Final(digest, &ctx);
+#endif
+
+    make_digest(resp, digest);
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): resp=%s", resp);
+#endif
+
+    sprintf(response, "charset=utf-8,username=\"%s\",realm=\"%s\",nonce=\"%s\","
+               "nc=00000001,cnonce=\"%s\",digest-uri=\"%s\",qop=%s,"
+               "response=%s",
+               login, realm, nonce, cnonce, uri, qop, resp);
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): response:%s", response);
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): text:%s", text);
+#endif
+}
+
+
 smtp_t *
 smtp_auth(config_t *cfg) {
 
@@ -348,6 +546,10 @@ smtp_auth(config_t *cfg) {
     }
     addr.sin_addr.s_addr = *((int *)he->h_addr_list[0]);
     s = socket(PF_INET, SOCK_STREAM, 0);
+
+    if(cfg->conn_timeout > 0) {
+        set_timeout(cfg->conn_timeout);
+    }
     if(connect(s, (struct sockaddr *) &addr, sizeof(addr))) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): connection error = %s",strerror(errno));
@@ -360,6 +562,9 @@ smtp_auth(config_t *cfg) {
         goto bail;
     }
     smtp->sock->fd = s;
+    if(cfg->conn_timeout > 0) {
+        set_timeout(0);
+    }
 
 #ifdef USE_SSL
     use_ssl = 0;
@@ -376,10 +581,13 @@ smtp_auth(config_t *cfg) {
 #endif
 
     /* CLAIM: we now have a TCP connection to the remote SMTP server */
-    alarm(cfg->timeout);
-
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = socket_read(smtp->sock, rbuf, sizeof(rbuf));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (banner): %m");
@@ -423,10 +631,14 @@ smtp_auth(config_t *cfg) {
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s%s", EHLO_CMD, myhostname);
 #endif
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = retry_writev(smtp->sock, iov, 3);
     memset(iov, 0, sizeof(iov));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): writev: %m");
@@ -439,9 +651,13 @@ smtp_auth(config_t *cfg) {
     }
 
     /* read and parse the EHLO response */
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = socket_read(smtp->sock, rbuf, sizeof(rbuf));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -557,10 +773,14 @@ smtp_quit(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", QUIT_CMD);
 #endif
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = retry_writev(sock, iov, 2);
     memset(iov, 0, sizeof(iov));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): quit writev: %m");
@@ -592,10 +812,14 @@ auth_cram_md5(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s%s", AUTH_CMD, "CRAM-MD5");
 #endif
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = retry_writev(sock, iov, 3);
     memset(iov, 0, sizeof(iov));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): cram-md5 writev: %m");
@@ -603,9 +827,13 @@ auth_cram_md5(socket_t *sock, config_t *cfg) {
         return AUTH_NG;
     }
 
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = socket_read(sock, rbuf, sizeof(rbuf));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -648,10 +876,14 @@ auth_cram_md5(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", response64);
 #endif
-        alarm(cfg->timeout);
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
         rc = retry_writev(sock, iov, 2);
         memset(iov, 0, sizeof(iov));
-        alarm(0);
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
         if(rc == -1) {
 #ifdef DEBUG
             syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): cram-md5 writev: %m");
@@ -659,9 +891,13 @@ auth_cram_md5(socket_t *sock, config_t *cfg) {
             return AUTH_NG;
         }
 
-        alarm(cfg->timeout);
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
         rc = socket_read(sock, rbuf, sizeof(rbuf));
-        alarm(0);
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
         if(rc == -1) {
 #ifdef DEBUG
             syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -711,10 +947,14 @@ auth_login(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s%s", AUTH_CMD, "LOGIN");
 #endif
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = retry_writev(sock, iov, 3);
     memset(iov, 0, sizeof(iov));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): login writev: %m");
@@ -722,9 +962,13 @@ auth_login(socket_t *sock, config_t *cfg) {
         return AUTH_NG;
     }
 
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = socket_read(sock, rbuf, sizeof(rbuf));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -747,10 +991,14 @@ auth_login(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", buf);
 #endif
-        alarm(cfg->timeout);
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
         rc = retry_writev(sock, iov, 2);
         memset(iov, 0, sizeof(iov));
-        alarm(0);
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
         if(rc == -1) {
 #ifdef DEBUG
             syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): login writev: %m");
@@ -758,9 +1006,13 @@ auth_login(socket_t *sock, config_t *cfg) {
             return AUTH_NG;
         }
 
-        alarm(cfg->timeout);
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
         rc = socket_read(sock, rbuf, sizeof(rbuf));
-        alarm(0);
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
         if(rc == -1) {
 #ifdef DEBUG
             syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -783,10 +1035,14 @@ auth_login(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
             syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", buf);
 #endif
-            alarm(cfg->timeout);
+            if(cfg->timeout > 0) {
+                set_timeout(cfg->timeout);
+            }
             rc = retry_writev(sock, iov, 2);
             memset(iov, 0, sizeof(iov));
-            alarm(0);
+            if(cfg->timeout > 0) {
+                set_timeout(0);
+            }
             if(rc == -1) {
 #ifdef DEBUG
                 syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): login writev: %m");
@@ -794,9 +1050,13 @@ auth_login(socket_t *sock, config_t *cfg) {
                 return AUTH_NG;
             }
 
-            alarm(cfg->timeout);
+            if(cfg->timeout > 0) {
+                set_timeout(cfg->timeout);
+            }
             rc = socket_read(sock, rbuf, sizeof(rbuf));
-            alarm(0);
+            if(cfg->timeout > 0) {
+                set_timeout(0);
+            }
             if(rc == -1) {
 #ifdef DEBUG
                 syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -865,11 +1125,15 @@ auth_plain(socket_t *sock, config_t *cfg) {
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s%s %s", AUTH_CMD, "PLAIN", buf);
 #endif
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = retry_writev(sock, iov, 4);
     memset(iov, 0, sizeof(iov));
     free(buf);
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): plain writev: %m");
@@ -877,9 +1141,13 @@ auth_plain(socket_t *sock, config_t *cfg) {
         return AUTH_NG;
     }
 
-    alarm(cfg->timeout);
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
     rc = socket_read(sock, rbuf, sizeof(rbuf));
-    alarm(0);
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
     if(rc == -1) {
 #ifdef DEBUG
         syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
@@ -901,11 +1169,195 @@ auth_plain(socket_t *sock, config_t *cfg) {
 int
 auth_digest_md5(socket_t *sock, config_t *cfg) {
 
+    struct iovec iov[3];
+    int rc;
+    char rbuf[RESP_LEN];
+    char *buf;
+
 #ifdef DEBUG
     syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): AUTH DIGEST-MD5");
 #endif
-    syslog(LOG_WARNING, "pam_smtpauth(smtpauth): auth_digest_md5 is not implemented.");
-    return AUTH_NG;
+
+    iov[0].iov_base = AUTH_CMD;
+    iov[0].iov_len  = sizeof(AUTH_CMD) - 1;
+    iov[1].iov_base = "DIGEST-MD5";
+    iov[1].iov_len  = strlen("DIGEST-MD5");
+    iov[2].iov_base = "\r\n";
+    iov[2].iov_len  = sizeof("\r\n") - 1;
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s%s", AUTH_CMD, "DIGEST-MD5");
+#endif
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
+    rc = retry_writev(sock, iov, 3);
+    memset(iov, 0, sizeof(iov));
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
+    if(rc == -1) {
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): digest-md5 writev: %m");
+#endif
+        return AUTH_NG;
+    }
+
+    if(cfg->timeout > 0) {
+        set_timeout(cfg->timeout);
+    }
+    rc = socket_read(sock, rbuf, sizeof(rbuf));
+    if(cfg->timeout > 0) {
+        set_timeout(0);
+    }
+    if(rc == -1) {
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
+#endif
+        return AUTH_NG;
+    }
+
+#ifdef DEBUG
+    syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %s",rbuf);
+#endif
+    if(strncmp(rbuf, "334 ", sizeof("334 ")-1) == 0) {
+        char *response;
+        char *response64;
+        char *challenge;
+        int challengelen;
+        unsigned char hexdigest[256];
+
+        challenge = malloc(strlen(rbuf + 4) + 1);
+        challengelen = base64_decode(challenge, rbuf + 4, -1);
+        challenge[challengelen] = '\0';
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): challenge=%s", challenge);
+#endif
+
+        digest_md5(hexdigest, challenge, challengelen, cfg->username, cfg->password);
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): hexdigest=%s", hexdigest);
+#endif
+
+        response = malloc(sizeof(char)*256);
+        snprintf(response, 256, "%s", hexdigest);
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): response=%s", response);
+#endif
+        response64 = malloc((strlen(response) + 3) * 2 + 1);
+        base64_encode(response64, response, strlen(response));
+        free(response);
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): response64=%s", response64);
+#endif
+
+        iov[0].iov_base = response64;
+        iov[0].iov_len  = strlen(response64);
+        iov[1].iov_base = "\r\n";
+        iov[1].iov_len  = sizeof("\r\n") - 1;
+
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", response64);
+#endif
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
+        rc = retry_writev(sock, iov, 2);
+        memset(iov, 0, sizeof(iov));
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
+        if(rc == -1) {
+#ifdef DEBUG
+            syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): digest-md5 writev: %m");
+#endif
+            return AUTH_NG;
+        }
+
+        if(cfg->timeout > 0) {
+            set_timeout(cfg->timeout);
+        }
+        rc = socket_read(sock, rbuf, sizeof(rbuf));
+        if(cfg->timeout > 0) {
+            set_timeout(0);
+        }
+        if(rc == -1) {
+#ifdef DEBUG
+            syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
+#endif
+            return AUTH_NG;
+        }
+
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %s",rbuf);
+#endif
+        if(strncmp(rbuf, "334 ", sizeof("334 ")-1) == 0) {
+            int buflen;
+
+            buf = malloc(strlen(rbuf + 4) + 1);
+            buflen = base64_decode(buf, rbuf + 4, -1);
+            buf[buflen] = '\0';
+
+            iov[0].iov_base = buf;
+            iov[0].iov_len  = strlen(buf);
+            iov[1].iov_base = "\r\n";
+            iov[1].iov_len  = sizeof("\r\n") - 1;
+
+#ifdef DEBUG
+            syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): sending %s", buf);
+#endif
+            if(cfg->timeout > 0) {
+                set_timeout(cfg->timeout);
+            }
+            rc = retry_writev(sock, iov, 2);
+            memset(iov, 0, sizeof(iov));
+            if(cfg->timeout > 0) {
+                set_timeout(0);
+            }
+            if(rc == -1) {
+#ifdef DEBUG
+                syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): digest-md5 writev: %m");
+#endif
+                return AUTH_NG;
+            }
+
+            if(cfg->timeout > 0) {
+                set_timeout(cfg->timeout);
+            }
+            rc = socket_read(sock, rbuf, sizeof(rbuf));
+            if(cfg->timeout > 0) {
+                set_timeout(0);
+            }
+            if(rc == -1) {
+#ifdef DEBUG
+                syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %m");
+#endif
+                return AUTH_NG;
+            }
+
+#ifdef DEBUG
+            syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): read (response): %s",rbuf);
+#endif
+            if(strncmp(rbuf, "235 ", sizeof("235 ")-1) != 0) {
+#ifdef DEBUG
+                syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): auth failure.");
+#endif
+                return AUTH_NG;
+            }
+        } else {
+#ifdef DEBUG
+            syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): it seems digest-md5 mech is not implemented.");
+#endif
+            return AUTH_NG;
+        }
+        free(response64);
+    } else {
+#ifdef DEBUG
+        syslog(LOG_DEBUG, "pam_smtpauth(smtpauth): it seems digest-md5 mech is not implemented.");
+#endif
+        return AUTH_NG;
+    }
+    return AUTH_OK;
 }
 
 #ifdef USE_SSL
